@@ -1,20 +1,19 @@
-from twisted.web.resource import Resource, IResource, getChildForRequest
-from twisted.web.iweb import IRenderable
-from twisted.web.template import flattenString
-from twisted.web import server
-
-from twisted.python import log
-
 from twisted.internet import defer
-
-
+from twisted.python import log, failure
+from twisted.web import server
+from twisted.web.iweb import IRenderable
+from twisted.web.resource import Resource, IResource, getChildForRequest
+from twisted.web.template import flattenString
 from werkzeug.exceptions import HTTPException
 
 from klein.interfaces import IKleinRequest
 
+
+
 __all__ = ["KleinResource", "ensure_utf8_bytes"]
 
 
+
 def ensure_utf8_bytes(v):
     """
     Coerces a value which is either a C{unicode} or C{str} to a C{str}.
@@ -25,6 +24,7 @@ def ensure_utf8_bytes(v):
     return v
 
 
+
 class _StandInResource(object):
     """
     A standin for a Resource.
@@ -34,6 +34,83 @@ class _StandInResource(object):
     """
 
 
+
+class _URLDecodeError(Exception):
+    """
+    Raised if one or more string parts of the URL could not be decoded.
+    """
+    __slots__ = ["errors"]
+
+    def __init__(self, errors):
+        """
+        @param errors: List of decoding errors.
+        @type errors: L{list} of L{tuple} of L{str},
+            L{twisted.python.failure.Failure}
+        """
+        self.errors = errors
+
+    def __repr__(self):
+        return "<URLDecodeError(errors={0!r})>".format(self.errors)
+
+
+
+def _extractURLparts(request):
+    """
+    Extracts and decodes URI parts from C{request}.
+
+    All strings must be UTF8-decodable.
+
+    @param request: A Twisted Web request.
+    @type request: L{twisted.web.iweb.IRequest}
+
+    @raise URLDecodeError: If one of the parts could not be decoded as UTF-8.
+
+    @return: L{tuple} of the URL scheme, the server name, the server port, the
+        path info and the script name.
+    @rtype: L{tuple} of L{unicode}, L{unicode}, L{int}, L{unicode}, L{unicode}
+    """
+    server_name = request.getRequestHostname()
+    server_port = request.getHost().port
+    if (bool(request.isSecure()), server_port) not in [
+            (True, 443), (False, 80)]:
+        server_name = '%s:%d' % (server_name, server_port)
+    script_name = ''
+    if request.prepath:
+        script_name = '/'.join(request.prepath)
+
+        if not script_name.startswith('/'):
+            script_name = '/' + script_name
+
+    path_info = ''
+    if request.postpath:
+        path_info = '/'.join(request.postpath)
+
+        if not path_info.startswith('/'):
+            path_info = '/' + path_info
+
+    url_scheme = u'https' if request.isSecure() else u'http'
+
+    utf8Failures = []
+    try:
+        server_name = server_name.decode("utf-8")
+    except UnicodeDecodeError:
+        utf8Failures.append(("SERVER_NAME", failure.Failure()))
+    try:
+        path_info = path_info.decode("utf-8")
+    except UnicodeDecodeError:
+        utf8Failures.append(("PATH_INFO", failure.Failure()))
+    try:
+        script_name = script_name.decode("utf-8")
+    except UnicodeDecodeError:
+        utf8Failures.append(("SCRIPT_NAME", failure.Failure()))
+
+    if utf8Failures:
+        raise _URLDecodeError(utf8Failures)
+
+    return url_scheme, server_name, server_port, path_info, script_name
+
+
+
 class KleinResource(Resource):
     """
     A ``Resource`` that can do URL routing.
@@ -46,31 +123,38 @@ class KleinResource(Resource):
         self._app = app
 
 
+    def __eq__(self, other):
+        if isinstance(other, KleinResource):
+            return vars(self) == vars(other)
+        return NotImplemented
+
+
+    def __ne__(self, other):
+        result = self.__eq__(other)
+        if result is NotImplemented:
+            return result
+        return not result
+
+
     def render(self, request):
         # Stuff we need to know for the mapper.
-        server_name = request.getRequestHostname()
-        server_port = request.getHost().port
-        if (bool(request.isSecure()), server_port) not in [
-                (True, 443), (False, 80)]:
-            server_name = '%s:%d' % (server_name, server_port)
-        script_name = ''
-        if request.prepath:
-            script_name = '/'.join(request.prepath)
+        try:
+            url_scheme, server_name, server_port, path_info, script_name = \
+                _extractURLparts(request)
+        except _URLDecodeError as e:
+            for what, fail in e.errors:
+                log.err(fail, "Invalid encoding in {what}.".format(what=what))
+            request.setResponseCode(400)
+            return b"Non-UTF-8 encoding in URL."
 
-            if not script_name.startswith('/'):
-                script_name = '/' + script_name
-
-        path_info = ''
-        if request.postpath:
-            path_info = '/'.join(request.postpath)
-
-            if not path_info.startswith('/'):
-                path_info = '/' + path_info
-
-        url_scheme = 'https' if request.isSecure() else 'http'
         # Bind our mapper.
-        mapper = self._app.url_map.bind(server_name, script_name, path_info=path_info,
-            default_method=request.method, url_scheme=url_scheme)
+        mapper = self._app.url_map.bind(
+            server_name,
+            script_name,
+            path_info=path_info,
+            default_method=request.method,
+            url_scheme=url_scheme,
+        )
         # Make the mapper available to the view.
         kleinRequest = IKleinRequest(request)
         kleinRequest.mapper = mapper
@@ -141,7 +225,7 @@ class KleinResource(Resource):
             # request is finished.
             if request_finished[0]:
                 if not failure.check(defer.CancelledError):
-                    log.err(failure, _why="Unhandled Error Processing Request.")
+                    log.err(failure, "Unhandled Error Processing Request.")
                 return
 
             # If there are no more registered handlers, apply some defaults
